/*
 * engine.c
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>

#include "global.h"
#include "utils.h"
#include "engine.h"


static pid_t pid;
static int toEngineFd;
static FILE *toEngine;
static int fromEngineFd;
static FILE *fromEngine;


static void sendToEngine(char *fmt, ...) {
  va_list ap;
  char buffer[GTP_BUFSIZE];

  va_start(ap, fmt);
  vsprintf(buffer, fmt, ap);
  va_end(ap);
  writeLog("engine <-- %s", buffer);
  fwrite(buffer, 1, strlen(buffer), toEngine);
  fwrite("\n", 1, 1, toEngine);
}


static void collectResponse(char *buffer, int maxSize) {
  char *p;
  int n;
  int newlines;
  int c;

  p = buffer;
  n = 0;
  newlines = 0;
  while (1) {
    c = fgetc(fromEngine);
    if (c == EOF) {
      error("pipe closed by engine");
    }
    if (c < ' ' && c != '\t' && c != '\n') {
      continue;
    }
    if (c == '\n') {
      newlines++;
      if (newlines == 2) {
        *p = '\0';
        return;
      }
    } else {
      newlines = 0;
    }
    if (c == '\t') {
      c = ' ';
    }
    if (n < maxSize - 1) {
      *p++ = c;
      n++;
    }
  }
}


static char *receiveFromEngine(char *buffer, int maxSize) {
  int n;
  char *p;

  collectResponse(buffer, maxSize);
  p = buffer;
  while (*p == ' ') {
    p++;
  }
  n = strlen(p);
  if (n > 0 && p[n - 1] == '\n') {
    p[n - 1] = '\0';
  }
  writeLog("engine --> %s", p);
  return p;
}


static void receiveOkFromEngine(void) {
  char buffer[GTP_BUFSIZE];
  char *p;

  p = receiveFromEngine(buffer, GTP_BUFSIZE);
  if (*p != '=') {
    error("cannot receive OK from engine");
  }
}


static int receiveIntegerFromEngine(void) {
  char buffer[GTP_BUFSIZE];
  char *p;

  p = receiveFromEngine(buffer, GTP_BUFSIZE);
  if (*p != '=') {
    error("cannot receive integer from engine");
  }
  p++;
  while (*p == ' ') {
    p++;
  }
  return atoi(p);
}


static char *receiveVertexFromEngine(void) {
  static char buffer[GTP_BUFSIZE];
  char *p;

  p = receiveFromEngine(buffer, GTP_BUFSIZE);
  if (*p != '=') {
    error("cannot receive vertex from engine");
  }
  p++;
  while (*p == ' ') {
    p++;
  }
  return p;
}


static int tokenize(char *str, char *tokens[], int maxTokens) {
  char *p;
  int n;

  p = strtok(str, " ");
  n = 0;
  while (p != NULL) {
    if (n < maxTokens) {
      tokens[n++] = p;
    }
    p = strtok(NULL, " ");
  }
  return n;
}


void engineStart(char *enginePath, char *engineArgs) {
  int pipeToEngine[2];
  int pipeFromEngine[2];
  char *argBuffer;
  char *args[MAX_ENGINE_ARGS];
  int n;

  if (pipe(pipeToEngine) < 0) {
    error("cannot establish pipe to engine");
  }
  if (pipe(pipeFromEngine) < 0) {
    error("cannot establish pipe from engine");
  }
  pid = fork();
  if (pid < 0) {
    error("cannot fork engine process");
  }
  if (pid == 0) {
    /* child (engine) */
    close(pipeToEngine[1]);
    close(pipeFromEngine[0]);
    close(0);
    dup(pipeToEngine[0]);
    close(pipeToEngine[0]);
    close(1);
    dup(pipeFromEngine[1]);
    close(pipeFromEngine[1]);
    /* we need a modifiable string for argument tokenization */
    argBuffer = malloc(strlen(engineArgs) + 1);
    if (argBuffer == NULL) {
      error("out of memory");
    }
    strcpy(argBuffer, engineArgs);
    args[0] = enginePath;
    n = 1 + tokenize(argBuffer, &args[1], MAX_ENGINE_ARGS - 2);
    args[n] = NULL;
    execvp(enginePath, args);
    error("cannot exec engine '%s'", enginePath);
  } else {
    /* parent (bridge) */
    close(pipeToEngine[0]);
    close(pipeFromEngine[1]);
    toEngineFd = pipeToEngine[1];
    toEngine = fdopen(toEngineFd, "w");
    if (toEngine == NULL) {
      error("cannot create file pointer for toEngineFd");
    }
    setlinebuf(toEngine);
    fromEngineFd = pipeFromEngine[0];
    fromEngine = fdopen(fromEngineFd, "r");
    if (fromEngine == NULL) {
      error("cannot create file pointer for fromEngineFd");
    }
    setlinebuf(fromEngine);
  }
}


void engineStop(void) {
  sendToEngine("quit");
  receiveOkFromEngine();
}


int engineGetProtocolVersion(void) {
  sendToEngine("protocol_version");
  return receiveIntegerFromEngine();
}


void engineSetBoardsize(int boardSize) {
  sendToEngine("boardsize %d", boardSize);
  receiveOkFromEngine();
}


void engineClearBoard(void) {
  sendToEngine("clear_board");
  receiveOkFromEngine();
}


void engineSetHandicap(int handicap) {
  if (handicap != 0) {
    sendToEngine("fixed_handicap %d", handicap);
    receiveOkFromEngine();
  }
}


void engineSetKomi(float komi) {
  sendToEngine("komi %.2f", komi);
  receiveOkFromEngine();
}


void engineSendMove(Color color, Cmd cmd) {
  if (cmd.type == CMD_PASS) {
    /* pass */
    sendToEngine("play %s pass", colorToString(color));
  } else
  if (cmd.type == CMD_MOVE) {
    /* ordinary move */
    sendToEngine("play %s %c%d",
                 colorToString(color),
                 numToColumn(cmd.x),
                 numToRow(cmd.y));
  } else {
    /* this should never happen */
    error("illegal command type in engineSendMove()");
  }
  receiveOkFromEngine();
}


void engineRequestMove(Color color) {
  sendToEngine("genmove %s", colorToString(color));
}


Cmd engineCheckForMove(void) {
  fd_set readfds;
  struct timeval timeout;
  Cmd cmd;
  char *vertex;

  FD_ZERO(&readfds);
  FD_SET(fromEngineFd, &readfds);
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  select(fromEngineFd + 1, &readfds, NULL, NULL, &timeout);
  if (!FD_ISSET(fromEngineFd, &readfds)) {
    cmd.type = CMD_NONE;
    return cmd;
  }
  vertex = receiveVertexFromEngine();
  if (strcasecmp(vertex, "resign") == 0) {
    cmd.type = CMD_RESIGN;
  } else
  if (strcasecmp(vertex, "pass") == 0) {
    cmd.type = CMD_PASS;
  } else {
    cmd.type = CMD_MOVE;
    cmd.x = columnToNum(vertex[0]);
    cmd.y = rowToNum(atoi(&vertex[1]));
  }
  return cmd;
}
