/*
 * main.c
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "global.h"
#include "utils.h"
#include "engine.h"
#include "serial.h"


static char *enginePath = DEFAULT_ENGINE_PATH;
static char *engineArgs = DEFAULT_ENGINE_ARGS;
static char *comDevice = DEFAULT_COM_DEVICE;
static int comBaudrate = DEFAULT_COM_BAUDRATE;
static int protocol = DEFAULT_PROTOCOL;
static int boardSize = DEFAULT_BOARD_SIZE;
static Color engineColor = DEFAULT_ENGINE_COLOR;
static int handicap = DEFAULT_HANDICAP;
static float komi = DEFAULT_KOMI;
static char *logPath = DEFAULT_LOG_PATH;

boolean debug = FALSE;


static void checkValuePresent(int argNum, int numArgs, char *arg) {
  if (argNum == numArgs - 1) {
    error("value for option '%s' missing", arg);
  }
}


static void help(char *myself) {
  printf("Usage: %s [options]\n", myself);
  printf("Options: (default values in parentheses)\n");
  printf("   --engine <path>          set Go engine path\n");
  printf("                            (%s)\n",
                                      DEFAULT_ENGINE_PATH);
  printf("   --args '<arguments>'     set Go engine arguments\n");
  printf("                            (%s)\n",
                                      DEFAULT_ENGINE_ARGS);
  printf("   --com <device>           set COM device\n");
  printf("                            (%s)\n",
                                      DEFAULT_COM_DEVICE);
  printf("   --baud <rate>            set COM baudrate\n");
  printf("                            (%d)\n",
                                      DEFAULT_COM_BAUDRATE);
  printf("   --protocol [gmp|sgmp]    set protocol variant\n");
  printf("                            (%s)\n",
                                      protocolToString(DEFAULT_PROTOCOL));
  printf("   --size <size>            set board size\n");
  printf("                            (%d)\n",
                                      DEFAULT_BOARD_SIZE);
  printf("   --color [w|b]            set color played by engine\n");
  printf("                            (%s)\n",
                                      colorToString(DEFAULT_ENGINE_COLOR));
  printf("   --handicap <n>           set handicap to n stones\n");
  printf("                            (%d)\n",
                                      DEFAULT_HANDICAP);
  printf("   --komi <f>               set komi to f points\n");
  printf("                            (%.2f)\n",
                                      DEFAULT_KOMI);
  printf("   --log <path>             set logfile path\n");
  printf("                            (%s)\n",
                                      DEFAULT_LOG_PATH);
  printf("   --debug                  enable debugging output\n");
  printf("                            (off)\n");
  printf("   --help                   show this help\n");
}


static void options(int argc, char *argv[]) {
  int i;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "--engine") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      enginePath = argv[i];
    } else
    if (strcmp(argv[i], "--args") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      engineArgs = argv[i];
    } else
    if (strcmp(argv[i], "--com") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      comDevice = argv[i];
    } else
    if (strcmp(argv[i], "--baud") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      comBaudrate = atoi(argv[i]);
      if (comBaudrate != 300 &&
          comBaudrate != 600 &&
          comBaudrate != 1200 &&
          comBaudrate != 2400 &&
          comBaudrate != 4800 &&
          comBaudrate != 9600 &&
          comBaudrate != 19200 &&
          comBaudrate != 38400) {
        error("supported baudrates are 300*2^n, n=0..7");
      }
    } else
    if (strcmp(argv[i], "--protocol") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      if (strcasecmp(argv[i], "gmp") == 0) {
        protocol = PROTOCOL_GMP;
      } else
      if (strcasecmp(argv[i], "sgmp") == 0) {
        protocol = PROTOCOL_SGMP;
      } else {
        error("unknown protocol '%s' specified", argv[i]);
      }
    } else
    if (strcmp(argv[i], "--size") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      boardSize = atoi(argv[i]);
    } else
    if (strcmp(argv[i], "--color") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      if (argv[i][0] == 'w' || argv[i][0] == 'W') {
        engineColor = WHITE;
      } else
      if (argv[i][0] == 'b' || argv[i][0] == 'B') {
        engineColor = BLACK;
      } else {
        error("illegal color '%s' specified", argv[i]);
      }
    } else
    if (strcmp(argv[i], "--handicap") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      handicap = atoi(argv[i]);
      if (handicap != 0 && !(handicap >= 2 && handicap <= 9)) {
        error("handicap must be in range 2..9, or 0");
      }
    } else
    if (strcmp(argv[i], "--komi") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      komi = atof(argv[i]);
    } else
    if (strcmp(argv[i], "--log") == 0) {
      checkValuePresent(i, argc, argv[i]);
      i++;
      logPath = argv[i];
    } else
    if (strcmp(argv[i], "--debug") == 0) {
      extern int gmp_debug;
      gmp_debug = 1;
      debug = TRUE;
    } else
    if (strcmp(argv[i], "--help") == 0) {
      help(argv[0]);
      exit(0);
    } else {
      error("unrecognized option '%s'; try '%s --help'",
            argv[i], argv[0]);
    }
  }
}


static void writeParameters(void) {
  writeLog("Program started");
  writeLog("  engine     %s", enginePath);
  writeLog("  args       %s", engineArgs);
  writeLog("  com        %s", comDevice);
  writeLog("  baud       %d", comBaudrate);
  writeLog("  protocol   %s", protocolToString(protocol));
  writeLog("  size       %d", boardSize);
  writeLog("  color      %s", colorToString(engineColor));
  writeLog("  handicap   %d", handicap);
  writeLog("  komi       %.2f", komi);
  writeLog("  log        %s", logPath);
}


int main(int argc, char *argv[]) {
  Cmd cmd;
  Color turn;
  int passes;

  options(argc, argv);
  openLog(logPath);
  writeParameters();
  engineStart(enginePath, engineArgs);
  if (engineGetProtocolVersion() != 2) {
    error("only GTP version 2 supported");
  }
  engineSetBoardsize(boardSize);
  engineClearBoard();
  engineSetHandicap(handicap);
  engineSetKomi(komi);
  serialInit(comDevice, comBaudrate);
  serialSendNewgame(boardSize, handicap, komi,
                    FALSE, engineColor == WHITE, protocol == PROTOCOL_SGMP);
  do {
    cmd = serialReceive(TRUE);
  } while (cmd.type == CMD_NONE || cmd.type == CMD_RESET);
  turn = (handicap == 0 ? BLACK : WHITE);
  passes = 0;
  while (1) {
    if (turn == engineColor) {
      /* our turn */
      engineRequestMove(turn);
      while (1) {
        cmd = engineCheckForMove();
        if (cmd.type != CMD_NONE) {
          /* the engine has something for us */
          break;
        }
        /* we have time to kick the serial connection */
        cmd = serialReceive(FALSE);
        if (cmd.type != CMD_NONE) {
          /* the GMP connection tries to get attention */
          error("untimely input from GMP connection");
        }
        /* let the engine work */
        sleep(1);
      }
      if (cmd.type == CMD_MOVE) {
        cmd.y = (boardSize - 1) - cmd.y;
        serialSendMove(cmd);
        passes = 0;
      } else
      if (cmd.type == CMD_PASS) {
        serialSendMove(cmd);
        if (++passes == 2) {
          /* game over */
          break;
        }
      } else
      if (cmd.type == CMD_RESIGN) {
        error("engine resigned");
      } else {
        /* this should never happen */
        error("illegal command type in main(), our turn");
      }
    } else {
      /* opponent's turn */
      do {
        cmd = serialReceive(TRUE);
      } while (cmd.type != CMD_MOVE &&
               cmd.type != CMD_PASS);
      if (cmd.type == CMD_MOVE) {
        cmd.y = (boardSize - 1) - cmd.y;
        engineSendMove(turn, cmd);
        passes = 0;
      } else
      if (cmd.type == CMD_PASS) {
        engineSendMove(turn, cmd);
        if (++passes == 2) {
          /* game over */
          break;
        }
      } else {
        /* this should never happen */
        error("illegal command type in main(), opponent's turn");
      }
    }
    turn = OPPONENT(turn);
  }
  engineStop();
  serialExit();
  closeLog();
  return 0;
}
