/*
 * serial.c
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "global.h"
#include "utils.h"
#include "serial.h"
#include "gmp.h"


static int serialLine;
static struct termios origSerialAttr;
static struct termios currSerialAttr;

static Gmp *ge;


static speed_t rateToSpeed(int rate) {
  switch (rate) {
    case 300:
      return B300;
    case 600:
      return B600;
    case 1200:
      return B1200;
    case 2400:
      return B2400;
    case 4800:
      return B4800;
    case 9600:
      return B9600;
    case 19200:
      return B19200;
    case 38400:
      return B38400;
  }
  /* this should never happen */
  error("illegal baudrate in rateToSpeed()");
  /* not reached */
  return 0;
}


void serialInit(char *comDevice, int comBaudrate) {
  serialLine = open(comDevice, O_RDWR | O_NDELAY);
  if (serialLine < 0) {
    error("cannot open COM device '%s' (%s)", comDevice, strerror(errno));
  }
  tcgetattr(serialLine, &origSerialAttr);
  currSerialAttr = origSerialAttr;
  currSerialAttr.c_cflag = rateToSpeed(comBaudrate) | CREAD | CLOCAL | CS8;
  currSerialAttr.c_iflag = 0;
  currSerialAttr.c_oflag = 0;
  currSerialAttr.c_lflag = 0;
  memset(currSerialAttr.c_cc, 0, NCCS);
  tcsetattr(serialLine, TCSANOW, &currSerialAttr);
  ge = gmp_create(serialLine, serialLine);
  if (ge == NULL) {
    error("cannot create GMP structure");
  }
}


void serialExit(void) {
  GmpResult result;
  int x, y;
  const char *err;

  sleep(5);
  result = gmp_check(ge, FALSE, &x, &y, &err);
  gmp_destroy(ge);
  tcsetattr(serialLine, TCSANOW, &origSerialAttr);
  close(serialLine);
}


void serialSendNewgame(int boardsize,
                       int handicap,
                       float komi,
                       int chineseRules,
                       int iAmWhite,
                       boolean simpleGmp) {
  writeLog("serial <-- newgame %d %d %.2f %d %d %d",
           boardsize, handicap, komi, chineseRules, iAmWhite, simpleGmp);
  gmp_startGame(ge, boardsize, handicap, komi,
                chineseRules, iAmWhite, simpleGmp);
}


void serialSendMove(Cmd cmd) {
  if (cmd.type == CMD_PASS) {
    /* pass */
    writeLog("serial <-- pass");
    gmp_sendPass(ge);
  } else
  if (cmd.type == CMD_MOVE) {
    /* ordinary move */
    writeLog("serial <-- move %d %d", cmd.x, cmd.y);
    gmp_sendMove(ge, cmd.x, cmd.y);
  } else {
    /* this should never happen */
    error("illegal command in serialSendMove()");
  }
}


Cmd serialReceive(boolean wait) {
  GmpResult result;
  int x, y;
  const char *err;
  Cmd cmd;

  result = gmp_check(ge, wait, &x, &y, &err);
  switch (result) {
    case gmp_nothing:
      cmd.type = CMD_NONE;
      break;
    case gmp_move:
      writeLog("serial --> move %d %d", x, y);
      cmd.type = CMD_MOVE;
      cmd.x = x;
      cmd.y = y;
      break;
    case gmp_pass:
      writeLog("serial --> pass");
      cmd.type = CMD_PASS;
      break;
    case gmp_reset:
      writeLog("serial --> reset");
      cmd.type = CMD_RESET;
      break;
    case gmp_newGame:
      writeLog("serial --> newgame");
      cmd.type = CMD_NEWGAME;
      break;
    case gmp_undo:
      writeLog("serial --> undo");
      cmd.type = CMD_UNDO;
      cmd.x = x;
      break;
    case gmp_err:
      writeLog("serial --> error");
      error("GMP failure (%s)", err);
      /* not reached */
      break;
  }
  return cmd;
}
